// Copyright (C) 2021 Toitware ApS. All rights reserved.
// Use of this source code is governed by an MIT-style license that can be
// found in the lib/LICENSE file.

import device
import rpc
import serialization show serialize deserialize
import uuid

RPC_SYSTEM_DEVICE_NAME ::= 100
RPC_SYSTEM_DEVICE_HARDWARE_ID ::= 101
RPC_SYSTEM_FLASHSTORE_GET ::= 102
RPC_SYSTEM_FLASHSTORE_SET ::= 103
RPC_SYSTEM_FLASHSTORE_DELETE ::= 104
RPC_SYSTEM_CONSOLE_CONNECTION_OPEN ::= 105
RPC_SYSTEM_CONSOLE_CONNECTION_CLOSE ::= 106
RPC_SYSTEM_ESTIMATE_TIME_ACCURACY ::= 107
RPC_SYSTEM_GNSS_START ::= 108
RPC_SYSTEM_GNSS_LOCATION ::= 109
RPC_SYSTEM_GNSS_STOP ::= 110

// TODO(kasper): Added late in v1.6.
RPC_SYSTEM_CONSOLE_CONNECTION_IS_OPEN ::= 111
RPC_SYSTEM_DEVICE_ICCID ::= 112
RPC_SYSTEM_CONSOLE_CONNECTION_AGO ::= 113
RPC_SYSTEM_CELLULAR_QUALITY ::= 114

class Device_ implements device.Device_:
  static instance_/Device_? := null

  constructor.instance:
    if not instance_: instance_ = Device_.init_

    return instance_

  constructor.init_:

  name -> string?:
    return rpc.invoke RPC_SYSTEM_DEVICE_NAME []

  iccid -> string?:
    return rpc.invoke RPC_SYSTEM_DEVICE_ICCID []

  cellular_signal_quality -> device.SignalQuality?:
    result := rpc.invoke RPC_SYSTEM_CELLULAR_QUALITY []
    if not result: return null
    return device.SignalQuality --power=result[0] --quality=result[1]

  hardware_id -> uuid.Uuid:
    return uuid.Uuid
      rpc.invoke RPC_SYSTEM_DEVICE_HARDWARE_ID []

  estimate_time_accuracy -> int?:
    return rpc.invoke RPC_SYSTEM_ESTIMATE_TIME_ACCURACY []

class FlashStore_ implements device.Store:
  static instance_/FlashStore_? := null

  constructor.instance:
    if not instance_: instance_ = FlashStore_.init_

    return instance_

  constructor.init_:

  get key/string -> any:
    res := rpc.invoke --sequential RPC_SYSTEM_FLASHSTORE_GET [key]
    if res is ByteArray:
      return deserialize res
    return null

  delete key/string:
    rpc.invoke --sequential RPC_SYSTEM_FLASHSTORE_DELETE [key]

  set key/string value/any:
    rpc.invoke --sequential RPC_SYSTEM_FLASHSTORE_SET [key, serialize value]

class ConsoleConnection_ extends rpc.CloseableProxy:
  constructor.open:
    super
      rpc.invoke RPC_SYSTEM_CONSOLE_CONNECTION_OPEN []

  static is_open -> bool:
    return rpc.invoke RPC_SYSTEM_CONSOLE_CONNECTION_IS_OPEN []

  static ago -> Duration?:
    ns/int? := rpc.invoke RPC_SYSTEM_CONSOLE_CONNECTION_AGO []
    return ns ? (Duration ns) : null

  close_rpc_selector_: return RPC_SYSTEM_CONSOLE_CONNECTION_CLOSE

class Gnss_ extends rpc.CloseableProxy:
  /**
  Starts GNSS.

  The device must support GNSS.
  */
  constructor.start:
    super
      rpc.invoke RPC_SYSTEM_GNSS_START []

  close_rpc_selector_: return RPC_SYSTEM_GNSS_STOP

  /**
  The location of this device.

  Returns null if there is no location fix.
  */
  location -> device.GnssLocation?:
    bytes := rpc.invoke RPC_SYSTEM_GNSS_LOCATION [handle_]
    if not bytes: return null
    return device.GnssLocation.deserialize bytes
